import{y as a,al as r,bG as n,bH as t,n as o,F as s,bI as e,ah as i,ag as u,af as f,aj as v}from"./index-CqPQBOGX.js";import{a as l}from"./_baseEach-BCfJ5kvl.js";import{d as c}from"./_baseUniq-PgmKRFB3.js";import{b}from"./_baseMap-D6WBeUOj.js";var m=Object.prototype,h=m.hasOwnProperty,d=a((function(a,o){a=Object(a);var s=-1,e=o.length,i=e>2?o[2]:void 0;for(i&&r(o[0],o[1],i)&&(e=1);++s<e;)for(var u=o[s],f=n(u),v=-1,l=f.length;++v<l;){var c=f[v],b=a[c];(void 0===b||t(b,m[c])&&!h.call(a,c))&&(a[c]=u[c])}return a}));function p(a){var r=null==a?0:a.length;return r?a[r-1]:void 0}var j=Math.max;var g,O=(g=function(a,r,n){var t=null==a?0:a.length;if(!t)return-1;var o=null==n?0:e(n);return o<0&&(o=j(t+o,0)),c(a,l(r),o)},function(a,r,n){var t=Object(a);if(!o(a)){var e=l(r);a=s(a),r=function(a){return e(t[a],a,t)}}var i=g(a,r,n);return i>-1?t[e?a[i]:i]:void 0});function x(a,r){return(i(a)?u:b)(a,l(r))}function y(a,r){return a<r}function _(a,r,n){for(var t=-1,o=a.length;++t<o;){var s=a[t],e=r(s);if(null!=e&&(void 0===i?e==e&&!f(e):n(e,i)))var i=e,u=s}return u}function M(a){return a&&a.length?_(a,v,y):void 0}export{y as a,_ as b,M as c,d,O as f,p as l,x as m};
