import{_ as r,l as s,G as t,j as a,H as e}from"./mermaid-ClBCjUJW.js";import{p as i}from"./gitGraph-YCYPL57B-CKQO5_4r.js";import"./index-CqPQBOGX.js";import"./transform-BZgIUWgO.js";import"./step-CjxFKgSe.js";import"./main-DPWVgc5Y.js";import"./_baseEach-BCfJ5kvl.js";import"./_baseUniq-PgmKRFB3.js";import"./min-BzJD3ah1.js";import"./_baseMap-D6WBeUOj.js";import"./clone-CoUwKDfY.js";var o={version:e},n={parser:{parse:r((async r=>{const t=await i("info",r);s.debug(t)}),"parse")},db:{getVersion:r((()=>o.version),"getVersion")},renderer:{draw:r(((r,e,i)=>{s.debug("rendering info diagram\n"+r);const o=t(e);a(o,100,400,!0);o.append("g").append("text").attr("x",100).attr("y",40).attr("class","version").attr("font-size",32).style("text-anchor","middle").text(`v${i}`)}),"draw")}};export{n as diagram};
