import{V as e}from"./VegaLite-Dkkjw4UL.js";import{a as r}from"./VegaLite-Dkkjw4UL.js";import{R as t}from"./index-CqPQBOGX.js";import"./compile-Cenw5ENu.js";import"./time-CgLoM9T8.js";import"./step-CjxFKgSe.js";import"./linear-HnjJmaH7.js";import"./init-DLRA0X12.js";import"./range-CtcPcB_L.js";import"./ordinal-BcaZTuz9.js";import"./colors-bszWmPJw.js";import"./arc-XdNuvsqn.js";import"./index-BbGmmqkJ.js";function o(){return o=Object.assign||function(e){for(var r=1;r<arguments.length;r++){var t=arguments[r];for(var o in t)Object.prototype.hasOwnProperty.call(t,o)&&(e[o]=t[o])}return e},o.apply(this,arguments)}function i(r){let{mode:i,spec:n}=r;class s extends t.PureComponent{render(){return t.createElement(e,o({mode:i,spec:n},this.props))}}var p,a,m;return m=function(){return n},(a="getSpec")in(p=s)?Object.defineProperty(p,a,{value:m,enumerable:!0,configurable:!0,writable:!0}):p[a]=m,s}export{e as Vega,r as VegaLite,i as createClassFromSpec};
